/********
 "sortspam" is a local delivery agent which is a wrapper around safecat
 (http://budney.homeunix.net:8080/users/budney/linux/software/safecat.html). It
 expects to receive a mail message on standard input, and then delivers the
 message to ./Mail/junkmail/ if a "X-Spam-Flag: YES" header is found, or
 ./Maildir/ otherwise. If a command line argument is given, it's text is used
 as the sorting header rather than "X-Spam-Flag: YES".
 
 Note that standard input must be able to be mmaped(*) for the sorting to
 succeed - this will be the case when qmail-local is doing the delivery.
 If mmap fails then the message is delivered to ./Maildir/.
 
 (*) mmap was used so that there are no local buffers used in the program, so
 that the program logic is very simple, and so that the file r/w pointer is not
 changed before safecat is run.
 ***********/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>
#include <uint32.h>
#include <errno.h>

char * find_str(const char* s, const long len, const char* pattern)
{
    char* sptr;
    long plen;
    long slen;
    char p;
    if (*pattern == 0) return 0;
    slen = len;
    plen = strlen(pattern);
    for (sptr = (char *)s; (p = *pattern) != 0; ++sptr, --slen)
    {
	if (slen <= plen)
	{
	    return 0;
	}
	if (!bcmp(sptr, pattern, strlen(pattern)))
	{
	    return sptr;
	}
    }
    return 0;
}

int main(int argc, char **argv)
{
    struct stat st;
    char *message;
    int fstat_ret;
    char *end_of_headers;
    char *spam_header = "X-Spam-Flag: YES";
    uint32 message_len;
    char* const good[] = { "safecat",
	"./Maildir/tmp",
	"./Maildir/new",
	0};
    char* const bad[] = { "safecat",
	"./Maildir/;junkmail/tmp",
	"./Maildir/;junkmail/new",
	0};

    if (argc > 1)
    {
	spam_header = argv[1];
    }
    // printf("Using %s as the spam header\n", spam_header);
    if (fstat_ret = fstat(0,&st))
    {
	fprintf(stderr,
		"Could not fstat standard input: %s\n", strerror(errno));
	exit(111);
    }
    message_len = (uint32)st.st_size;
    if (message_len > 0xffffffff)
    {
	fprintf(stderr, "Input file too big - %x\n", message_len);
	exit(100);
    }
    message = mmap(0,message_len,PROT_READ,MAP_SHARED,0,0);
    if (message + 1)
    {
	//printf("MMap succeeded, filesize %d pointer %x\n",
	//message_len, message);
	if (end_of_headers = find_str(message, message_len, "\n\n"))
	{
	// printf("Found end of headers at %x\n", end_of_headers);
	}
 	if (find_str(message, end_of_headers - message + 1, spam_header))
	{
	// printf("Found spam marker header\n");
	    munmap(message, message_len);
 	    fprintf(stderr, "Sorted as spam:");
	    execv("/usr/bin/safecat", bad);
 	    fprintf(stderr,
		    "Could not execute safecat: %s\n", strerror(errno));
 	    exit(111);
	}
	munmap(message, message_len);
    }
    else
    {
	//printf("MMap failed\n");
    }
    execv("/usr/bin/safecat", good);
    fprintf(stderr, "Could not execute safecat: %s\n", strerror(errno));
    exit(111);
}
